'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = convertTypeToPropTypes;

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _babelLog = require('babel-log');

var _babelFileLoader = require('babel-file-loader');

var _matchExported = require('./matchExported');

var _matchExported2 = _interopRequireDefault(_matchExported);

var _babelErrors = require('babel-errors');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function error(path, message) {
  var err = (0, _babelErrors.buildCodeFrameError)(path, message);
  err.stack = (0, _babelErrors.toErrorStack)(err);
  return err;
}

function cloneComments(comments) {
  return comments && comments.map(function (comment) {
    comment = t.clone(comment);
    comment.start = comment.start + 0.0001; // Force printer to print... (sigh)
    return comment;
  });
}

function inheritsComments(a, b) {
  return t.inheritsComments(a, {
    trailingComments: cloneComments(b.trailingComments),
    leadingComments: cloneComments(b.leadingComments),
    innerComments: cloneComments(b.innerComments)
  });
}

var refPropTypes = function refPropTypes(property, opts) {
  return t.memberExpression(opts.propTypesRef, property);
};

var createThrows = function createThrows(message) {
  return function (path, opts) {
    throw error(path, message);
  };
};

var createConversion = function createConversion(name) {
  return function (path, opts) {
    return refPropTypes(t.identifier(name), opts);
  };
};

var converters = {};

converters.AnyTypeAnnotation = createConversion('any');
converters.MixedTypeAnnotation = createConversion('any');
converters.NumberTypeAnnotation = createConversion('number');
converters.NumericLiteralTypeAnnotation = createConversion('number');
converters.BooleanTypeAnnotation = createConversion('bool');
converters.BooleanLiteralTypeAnnotation = createConversion('bool');
converters.StringTypeAnnotation = createConversion('string');
converters.StringLiteralTypeAnnotation = createConversion('string');
converters.NullLiteralTypeAnnotation = createThrows('null types unsupported');
converters.VoidTypeAnnotation = createThrows('void types unsupported');
converters.FunctionTypeAnnotation = createConversion('func');
converters.NullableTypeAnnotation = createThrows('maybe types unsupported');
converters.TupleTypeAnnotation = createConversion('array');

converters.QualifiedTypeIdentifier = createThrows('qualified type identifiers unsupported');

converters.TypeAnnotation = function (path, opts) {
  return convert(path.get('typeAnnotation'), opts);
};

converters.ObjectTypeAnnotation = function (path, opts) {
  var properties = [];

  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = path.get('properties')[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var property = _step.value;

      properties.push(convert(property, opts));
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  if (path.node.indexers.length) {
    throw error(path.get('indexers')[0], 'Object indexers are unsupported');
  }

  if (path.node.callProperties.length) {
    throw error(path.get('callProperties')[0], 'Object call properties are unsupported');
  }

  var object = t.objectExpression(properties);

  return t.callExpression(refPropTypes(t.identifier('shape'), opts), [object]);
};

converters.ObjectTypeProperty = function (path, opts) {
  var key = path.get('key');
  var value = path.get('value');

  var id = void 0;
  if (key.isStringLiteral()) {
    id = t.stringLiteral(key.node.value);
  } else {
    id = t.identifier(key.node.name);
  }

  var converted = convert(value, opts);

  if (!path.node.optional) {
    converted = t.memberExpression(converted, t.identifier('isRequired'));
  }

  return t.objectProperty(inheritsComments(id, key.node), converted);
};

converters.ArrayTypeAnnotation = function (path, opts) {
  return t.callExpression(refPropTypes(t.identifier('arrayOf'), opts), [convert(path.get('elementType'), opts)]);
};

var typeParametersConverters = {
  Array: function Array(path, opts) {
    var param = path.get('typeParameters').get('params')[0];
    return t.callExpression(refPropTypes(t.identifier('arrayOf'), opts), [convert(param, opts)]);
  }
};

converters.GenericTypeAnnotation = function (path, opts) {
  if (!path.node.typeParameters) {
    return convert(path.get('id'), opts);
  }

  var name = path.node.id.name;

  if (typeParametersConverters[name]) {
    return typeParametersConverters[name](path, opts);
  } else {
    throw error(path, 'Unsupported generic type annotation with type parameters');
  }
};

var typeIdentifierConverters = {
  Function: createConversion('func'),
  Object: createConversion('object')
};

converters.Identifier = function (path, opts) {
  var name = path.node.name;

  if (path.parentPath.isFlow() && typeIdentifierConverters[name]) {
    return typeIdentifierConverters[name](path, opts);
  }

  var binding = path.scope.getBinding(name);
  if (!binding) {
    throw error(path, 'Missing reference "' + name + '"');
  }
  return convert(binding.path, opts);
};

converters.TypeAlias = function (path, opts) {
  return convert(path.get('right'), opts);
};

converters.InterfaceDeclaration = function (path, opts) {
  return convert(path.get('body'), opts);
};

converters.ClassDeclaration = function (path, opts, id) {
  return t.callExpression(refPropTypes(t.identifier('instanceOf'), opts), [id || t.identifier(path.node.id.name)]);
};

converters.UnionTypeAnnotation = function (path, opts) {
  var types = path.get('types').map(function (p) {
    return convert(p, opts);
  });
  var arr = t.arrayExpression(types);

  return t.callExpression(refPropTypes(t.identifier('oneOf'), opts), [arr]);
};

function _convertImportSpecifier(path, opts) {
  var kind = path.parent.importKind;
  if (kind === 'typeof') {
    throw error(path, 'import typeof is unsupported');
  }

  var file = (0, _babelFileLoader.loadImportSync)(path.parentPath, opts.resolveOpts);
  var local = path.node.local.name;
  var name = void 0;

  if (path.type === 'ImportDefaultSpecifier' && kind === 'value') {
    name = 'default';
  } else {
    name = local;
  }

  var exported = (0, _matchExported2.default)(file, name);

  if (!exported) {
    throw error(path, 'Missing matching export');
  }

  return convert(exported, opts, t.identifier(name));
}

converters.ImportDefaultSpecifier = function (path, opts) {
  return _convertImportSpecifier(path, opts);
};

converters.ImportSpecifier = function (path, opts) {
  return _convertImportSpecifier(path, opts);
};

var convert = function convert(path, opts, id) {
  var converter = converters[path.type];

  if (!converter) {
    throw error(path, 'No converter for node type: ' + path.type);
  }

  return inheritsComments(converter(path, opts, id), path.node);
};

function convertTypeToPropTypes(typeAnnotation, propTypesRef, resolveOpts) {
  return convert(typeAnnotation, { propTypesRef: propTypesRef, resolveOpts: resolveOpts }).arguments[0];
}